import org.ejml.simple.SimpleMatrix;

import java.util.ArrayList;
import java.util.function.Function;

public class Network {

    ArrayList<Layer> layers;
    Function<SimpleMatrix, Double> loss;
    Function<SimpleMatrix, Double> lossPrime;

    public Network() {
        layers = new ArrayList<>();
    }

    public void add(Layer layer) {
        layers.add(layer);
    }

    public void use(Function<SimpleMatrix, Double> loss, Function<SimpleMatrix, Double> lossPrime) {
        this.loss = loss;
        this.lossPrime = lossPrime;
    }

    public ArrayList<SimpleMatrix> predict(SimpleMatrix[] inputs) {
        int samples = inputs.length;
        ArrayList<SimpleMatrix> result = new ArrayList<>();
        SimpleMatrix output;

        for (SimpleMatrix input : inputs) {
            output = input;
            for (Layer l : layers) {
                output = l.forwardPropagation(output);
            }
            result.add(output);
        }

        return result;
    }

    public void fit(SimpleMatrix[] X_train, SimpleMatrix[] y_train, int epochs, double learningRate) {
        int samples = X_train.length;

        for (int i = 0; i < epochs; i++) {
            double err = 0;
            for (int j = 0; j < samples; j++) {
                // forward propagation
                SimpleMatrix output = X_train[j];
                for (Layer l : layers) {
                    output = l.forwardPropagation(output);
                }

                // compute loss (for display purpose only)
                err = MSE(y_train[j], output);

                // backward propagation
                SimpleMatrix error = MSEPrime(y_train[j], output);
                for (int k = layers.size() - 1; k >= 0; k--) {
                    error = layers.get(k).backwardPropagation(error, learningRate);
                }
            }
            // calculate average error on all samples
            err /= samples;
            System.out.println("epoch " + (i+1) + "/" + epochs + "  error=" + err);
        }
    }

    private double MSE(SimpleMatrix y_true, SimpleMatrix y_pred) {
        SimpleMatrix temp = y_true.minus(y_pred);
        temp = temp.elementMult(temp);
        double sum = 0;
        for (int i = 0; i < temp.getNumElements(); i++) {
            sum += temp.get(i);
        }
        return sum / temp.getNumElements();
    }

    private SimpleMatrix MSEPrime(SimpleMatrix y_true, SimpleMatrix y_pred) {
        return y_true.minus(y_pred).divide((double) y_true.getNumElements()/2);
    }
}